@@ -296,6 +296,8 @@ PHP_MINIT_FUNCTION(rhash)
 	REGISTER_RHASH_CONSTANT(RHASH_SHA3_512);
 	REGISTER_RHASH_CONSTANT(RHASH_SNEFRU128);
 	REGISTER_RHASH_CONSTANT(RHASH_SNEFRU256);
+	REGISTER_RHASH_CONSTANT(RHASH_BLAKE2S);
+	REGISTER_RHASH_CONSTANT(RHASH_BLAKE2B);
 	REGISTER_RHASH_CONSTANT(RHASH_ALL);
 
 	return SUCCESS;
