@@ -11,7 +11,7 @@ extern "C" {
 #endif
 
 /**
- * Information about a file to calculate/verify hashes for.
+ * Information about a file to calculate/verify message digests for.
  */
 struct file_info {
 	uint64_t size;          /* the size of the hashed file */
@@ -19,15 +19,16 @@ struct file_info {
 	double time;            /* file processing time in seconds */
 	file_t* file;           /* the file being processed */
 	struct rhash_context* rctx; /* state of hash algorithms */
+	struct hash_parser* hp;  /* parsed line of a hash file */
+	unsigned sums_flags;    /* mask of ids of calculated hash functions */
 	int processing_result;  /* -1/-2 for i/o error, 0 on success, 1 on a hash mismatch */
-	unsigned sums_flags; /* mask of ids of calculated hash functions */
-	struct hash_check hc; /* hash values parsed from a hash file */
 };
 
-int save_torrent_to(file_t* torrent_file, struct rhash_context* rctx);
+int calc_sums(struct file_info* info);
 int calculate_and_print_sums(FILE* out, file_t* out_file, file_t* file);
-int check_hash_file(file_t* file, int chdir);
+int find_embedded_crc32(file_t* file, unsigned* crc32);
 int rename_file_by_embeding_crc32(struct file_info* info);
+int save_torrent_to(file_t* torrent_file, struct rhash_context* rctx);
 
 /* Benchmarking */
 
