@@ -14,76 +14,76 @@
 # define _(str) (str)
 #endif /* USE_GETTEXT */
 
+#include <stddef.h> /* for wchar_t */
 #include <stdint.h>
 #include <stdio.h>
 #include <time.h> /* for time_t */
-#include <stddef.h> /* for wchar_t */
-
 #if !defined( _WIN32) && !defined(__CYGWIN__)
 # include <sys/time.h> /* for timeval */
 #elif _MSC_VER > 1300
 # include "platform.h"
 #endif
 
-
 #ifdef __cplusplus
 extern "C" {
 #endif
 
 /* string function */
-void sprintI64(char *dst, uint64_t number, int max_width);
+void sprintI64(char* dst, uint64_t number, int max_width);
 int  int_len(uint64_t num);
 
-int  urlencode(char *dst, const char *name);
+int  urlencode(char* dst, const char* name);
+size_t count_utf8_symbols(const char* str);
 int  is_binary_string(const char* str);
 char* str_tolower(const char* str);
 char* str_trim(char* str);
 char* str_set(char* buf, int ch, int size);
-char* str_append(const char* orig, const char* append);
-size_t strlen_utf8_c(const char *str);
+char* str_replace_n(const char* src, size_t start_pos, size_t end_pos, const char* replace);
+#ifdef _WIN32
+wchar_t* wcs_replace_n(const wchar_t* src, size_t start_pos, size_t end_pos, const char* replace);
+#endif /* _WIN32 */
 
+/* check if character starts a commentary in the program config file */
 #define IS_COMMENT(c) ((c) == ';' || (c) == '#')
+/* test line for UTF8 mark: 0xEF, 0xBB, 0xBF */
+#define STARTS_WITH_UTF8_BOM(line) ((line)[0] == '\357' && (line)[1] == '\273' && (line)[2] == '\277')
 
 #ifdef _WIN32
 typedef wchar_t rsh_tchar;
 # define RSH_T(str) L##str
-# define t2c(tstr) (w2c(tstr))
 #else
 typedef  char rsh_tchar;
 # define RSH_T(str) str
-# define t2c(tstr) (tstr)
 #endif /* _WIN32 */
 typedef rsh_tchar* tstr_t;
 typedef const rsh_tchar* ctstr_t;
 
 #ifdef _WIN32
 # define IF_WINDOWS(code) code
-# define is_utf8() win_is_utf8()
-# define to_utf8(str) str_to_utf8(str)
 #else /* non _WIN32 part */
 # define IF_WINDOWS(code)
-/* stub for utf8 */
-# define is_utf8() 1
-# define to_utf8(str) NULL
 #endif /* _WIN32 */
 
 /* version information */
 const char* get_version_string(void);
 const char* get_bt_program_name(void);
 
-
+/* printf functions */
 #ifdef _WIN32
 # define rsh_fprintf win_fprintf
+# define rsh_fprintf_targ win_fprintf_warg
 # define rsh_vfprintf win_vfprintf
 # define rsh_fwrite win_fwrite
 #else
 # define rsh_fprintf fprintf
+# define rsh_fprintf_targ fprintf
 # define rsh_vfprintf vfprintf
 # define rsh_fwrite fwrite
 #endif
 
+#define PRINTF_RES(res) ((res) < 0 ? res : 0)
 
-/* time data and functions */
+/* time delta and timer functions */
 
 /* portable timer definition */
 #if defined( _WIN32) || defined(__CYGWIN__)
@@ -121,6 +121,7 @@ unsigned rhash_get_ticks(void);
 typedef void (*exit_handler_t)(void);
 void rsh_install_exit_handler(exit_handler_t handler);
 void rsh_remove_exit_handler(void);
+void rsh_call_exit_handlers(void);
 void rsh_exit(int code);
 
 /* clever malloc with error detection */
@@ -135,7 +136,10 @@ void* rhash_realloc(void* mem, size_t size, const char* srcfile, int srcline);
 
 #ifdef _WIN32
 #define rsh_wcsdup(str) rhash_wcsdup(str, __FILE__, __LINE__)
+#define rsh_tstrdup(str) rsh_wcsdup(str)
 wchar_t* rhash_wcsdup(const wchar_t* str, const char* srcfile, int srcline);
+#else
+#define rsh_tstrdup(str) rsh_strdup(str)
 #endif
 
 extern void (*rsh_report_error)(const char* srcfile, int srcline, const char* format, ...);
@@ -143,7 +147,7 @@ extern void (*rsh_report_error)(const char* srcfile, int srcline, const char* fo
 /* vector functions */
 typedef struct vector_t
 {
-	void **array;
+	void** array;
 	size_t size;
 	size_t allocated;
 	void (*destructor)(void*);
@@ -153,7 +157,7 @@ vector_t* rsh_vector_new(void (*destructor)(void*));
 vector_t* rsh_vector_new_simple(void);
 void rsh_vector_free(vector_t* vect);
 void rsh_vector_destroy(vector_t* vect);
-void rsh_vector_add_ptr(vector_t* vect, void *item);
+void rsh_vector_add_ptr(vector_t* vect, void* item);
 void rsh_vector_add_empty(vector_t* vect, size_t item_size);
 #define rsh_vector_add_uint32(vect, item) { \
 	rsh_vector_add_empty(vect, item_size); \
@@ -204,9 +208,9 @@ typedef struct strbuf_t
 
 strbuf_t* rsh_str_new(void);
 void rsh_str_free(strbuf_t* buf);
-void rsh_str_ensure_size(strbuf_t *str, size_t new_size);
-void rsh_str_append_n(strbuf_t *str, const char* text, size_t len);
-void rsh_str_append(strbuf_t *str, const char* text);
+void rsh_str_ensure_size(strbuf_t* str, size_t new_size);
+void rsh_str_append_n(strbuf_t* str, const char* text, size_t len);
+void rsh_str_append(strbuf_t* str, const char* text);
 
 #define rsh_str_ensure_length(str, len) \
 	if ((size_t)(len) >= (size_t)(str)->allocated) rsh_str_ensure_size((str), (len) + 1);
