@@ -17,13 +17,13 @@
 #include <stdlib.h>
 #include <string.h>
 #include <sys/stat.h>
+#include <fcntl.h>  /* _O_RDONLY, _O_BINARY, posix_fadvise */
 
-#if defined( _WIN32) || defined(__CYGWIN__)
+#if defined(_WIN32) || defined(__CYGWIN__)
 # include <windows.h>
 #if !defined(__CYGWIN__)
 # include <share.h> /* for _SH_DENYWR */
 #endif
-# include <fcntl.h>  /* _O_RDONLY, _O_BINARY */
 # include <io.h>
 #endif
 
@@ -62,8 +62,8 @@ static const char* get_basename(const char* path)
 	const char* p;
 	if (!path)
 		return NULL;
-	for (p = path + strlen(path) - 1; p >= path && !IS_PATH_SEPARATOR(*p); p--);
-	return (p + 1);
+	for (p = path + strlen(path); p > path && !IS_PATH_SEPARATOR(*(p - 1)); p--);
+	return p;
 }
 
 /**
@@ -184,6 +184,22 @@ int are_paths_equal(ctpath_t path, file_t* file)
 	return (*path == *fpath);
 }
 
+#ifndef _WIN32
+/**
+ * Convert a windows file path to a UNIX one, replacing '\\' by '/'.
+ *
+ * @param path the path to convert
+ * @return converted path
+ */
+static void convert_backslashes_to_unix(char* path)
+{
+	for (; *path; path++) {
+		if (*path == '\\')
+			*path = '/';
+	}
+}
+#endif /* _WIN32 */
+
 /**
  * Check if a path points to a regular file.
  *
@@ -226,7 +242,7 @@ enum FileMemoryModeBits {
  */
 int file_init(file_t* file, ctpath_t path, unsigned init_flags)
 {
-#if _WIN32
+#ifdef _WIN32
 	tpath_t long_path = get_long_path_if_needed(path);
 #endif
 	memset(file, 0, sizeof(*file));
@@ -243,7 +259,7 @@ int file_init(file_t* file, ctpath_t path, unsigned init_flags)
 			return -1;
 		}
 	}
-#if _WIN32
+#ifdef _WIN32
 	if (long_path)
 	{
 		file->real_path = long_path;
@@ -254,8 +270,12 @@ int file_init(file_t* file, ctpath_t path, unsigned init_flags)
 	{
 		if ((init_flags & FileInitReusePath) == 0)
 		{
-			file->real_path = rsh_tstrdup(path);
 			file->mode = init_flags & FileMaskModeBits;
+			file->real_path = rsh_tstrdup(path);
+#ifndef _WIN32
+			if ((init_flags & FileInitUseRealPathAsIs) == 0)
+				convert_backslashes_to_unix(file->real_path);
+#endif
 		}
 	}
 	if ((init_flags & (FileInitRunFstat | FileInitRunLstat)) &&
@@ -742,6 +762,7 @@ FILE* file_fopen(file_t* file, int fopen_flags)
 	const file_tchar* possible_modes[8] = { 0, RSH_T("r"), RSH_T("w"), RSH_T("r+"),
 		0, RSH_T("rb"), RSH_T("wb"), RSH_T("r+b") };
 	const file_tchar* mode = possible_modes[fopen_flags & FOpenMask];
+	FILE* fd;
 	assert((fopen_flags & FOpenRW) != 0);
 	if (!file->real_path) {
 		errno = EINVAL;
@@ -749,13 +770,18 @@ FILE* file_fopen(file_t* file, int fopen_flags)
 	}
 #ifdef _WIN32
 	{
-		FILE* fd = _wfsopen(file->real_path, mode, _SH_DENYNO);
+		fd = _wfsopen(file->real_path, mode, _SH_DENYNO);
 		if (!fd && errno == EINVAL)
 			errno = ENOENT;
 		return fd;
 	}
 #else
-	return fopen(file->real_path, mode);
+	fd = fopen(file->real_path, mode);
+# if _POSIX_C_SOURCE >= 200112L && !defined(__STRICT_ANSI__)
+	if(fd)
+		posix_fadvise(fileno(fd), 0, 0, POSIX_FADV_SEQUENTIAL);
+# endif /* _POSIX_C_SOURCE >= 200112L && !defined(__STRICT_ANSI__) */
+	return fd;
 #endif
 }
 
