@@ -1,4 +1,4 @@
-/* hash_print.h - functions to print hash sums */
+/* hash_print.h - functions to print message digests */
 #ifndef HASH_PRINT_H
 #define HASH_PRINT_H
 
@@ -26,6 +26,7 @@ typedef struct print_item
 typedef struct print_hash_info
 {
 	const char* name;
+	const char* bsd_name;
 	char short_name[20];
 	char short_char;
 } print_hash_info;
@@ -40,7 +41,7 @@ struct strbuf_t;
 void init_hash_info_table(void);
 void init_printf_format(struct strbuf_t* out);
 
-/* formatted output of hash sums and file information */
+/* formatted output of message digests and file information */
 print_item* parse_print_string(const char* format, unsigned* sum_mask);
 int print_line(FILE* out, unsigned out_mode, print_item* list, struct file_info* info);
 void free_print_list(print_item* list);
